@@ -135,15 +135,16 @@ public class MethodInjector {
 
     private static class WithBridgeMethodsAnnotationVisitor extends AnnotationVisitor {
       protected boolean castRequired = false;
+      protected String adapterMethod = null;
       protected List<Type> types = new ArrayList<Type>();
       
       public WithBridgeMethodsAnnotationVisitor(AnnotationVisitor av) {
-        super(Opcodes.ASM4, av);
+        super(Opcodes.ASM9, av);
       }
 
       @Override
       public AnnotationVisitor visitArray(String name) {
-        return new AnnotationVisitor(Opcodes.ASM4, super.visitArray(name)) {
+        return new AnnotationVisitor(Opcodes.ASM9, super.visitArray(name)) {
            
             public void visit(String name, Object value) {
                 if (value instanceof Type) {
@@ -161,6 +162,9 @@ public class MethodInjector {
         if ("castRequired".equals(name) && value instanceof Boolean) {
           castRequired = (Boolean) value;
         }
+        if ("adapterMethod".equals(name) && value instanceof String) {
+          adapterMethod = (String) value;
+        }
         super.visit(name, value);
       }
     }
@@ -179,10 +183,18 @@ public class MethodInjector {
             final String originalSignature;
             final String[] exceptions;
             final boolean castRequired;
+            final String adapterMethod;
 
+            /**
+             * Return type of the bridge method to be inserted.
+             */
             final Type returnType;
+            /**
+             * Return type of the declared method written in the source code.
+             */
+            final Type originalReturnType;
 
-            SyntheticMethod(int access, String name, String desc, String originalSignature, String[] exceptions, Type returnType, boolean castRequired) {
+            SyntheticMethod(int access, String name, String desc, String originalSignature, String[] exceptions, Type returnType, boolean castRequired, String adapterMethod) {
                 this.access = access;
                 this.name = name;
                 this.desc = desc;
@@ -190,6 +202,8 @@ public class MethodInjector {
                 this.exceptions = exceptions;
                 this.returnType = returnType;
                 this.castRequired = castRequired;
+                this.adapterMethod = adapterMethod;
+                originalReturnType = Type.getReturnType(desc);
             }
 
             /**
@@ -197,7 +211,6 @@ public class MethodInjector {
              */
             public void inject(ClassVisitor cv) {
                 Type[] paramTypes = Type.getArgumentTypes(desc);
-                Type originalReturnType = Type.getReturnType(desc);
 
                 int access = this.access | ACC_SYNTHETIC | ACC_BRIDGE;
                 String methodDescriptor = Type.getMethodDescriptor(returnType, paramTypes);
@@ -206,11 +219,19 @@ public class MethodInjector {
                 if ((access&ACC_ABSTRACT)==0) {
                     GeneratorAdapter ga = new GeneratorAdapter(mv, access, name, methodDescriptor);
                     mv.visitCode();
+
                     int sz = 0;
+
+                    if (hasAdapterMethod()) {
+                        // the LHS of the adapter method invocation
+                        ga.loadThis();
+                        sz++;
+                    }
+
                     boolean isStatic = (access & ACC_STATIC) != 0;
                     if (!isStatic) {
-                      mv.visitVarInsn(ALOAD,0);
-                      sz++;
+                        ga.loadThis();
+                        sz++;
                     }
 
                     for (Type p : paramTypes) {
@@ -219,6 +240,9 @@ public class MethodInjector {
                     }
                     mv.visitMethodInsn(
                       isStatic ? INVOKESTATIC : INVOKEVIRTUAL,internalClassName,name,desc);
+                    if (hasAdapterMethod()) {
+                        insertAdapterMethod(ga);
+                    } else
                     if (castRequired) {
                         ga.unbox(returnType);
                     } else {
@@ -242,10 +266,25 @@ public class MethodInjector {
                 }
                 mv.visitEnd();
             }
+
+            private boolean hasAdapterMethod() {
+                return adapterMethod!=null && adapterMethod.length()>0;
+            }
+
+            private void insertAdapterMethod(GeneratorAdapter ga) {
+                ga.push(returnType);
+                ga.visitMethodInsn(INVOKEVIRTUAL, internalClassName, adapterMethod,
+                        Type.getMethodDescriptor(
+                                Type.getType(Object.class), // return type
+                                originalReturnType,
+                                Type.getType(Class.class)
+                        ));
+                ga.unbox(returnType);
+            }
         }
 
         Transformer(ClassVisitor cv) {
-            super(Opcodes.ASM4, cv);
+            super(Opcodes.ASM9, cv);
         }
 
         @Override
@@ -267,7 +306,7 @@ public class MethodInjector {
         @Override
         public MethodVisitor visitMethod(final int access, final String name, final String mdesc, final String signature, final String[] exceptions) {
             MethodVisitor mv = super.visitMethod(access, name, mdesc, signature, exceptions);
-            return new MethodVisitor(Opcodes.ASM4, mv) {
+            return new MethodVisitor(Opcodes.ASM9, mv) {
                 @Override
                 public AnnotationVisitor visitAnnotation(String adesc, boolean visible) {
                     AnnotationVisitor av = super.visitAnnotation(adesc, visible);
@@ -279,7 +318,7 @@ public class MethodInjector {
                                 super.visitEnd(); 
                                 for (Type type : this.types)
                                     syntheticMethods.add(new SyntheticMethod(
-                                         access,name,mdesc,signature,exceptions,type, this.castRequired
+                                         access,name,mdesc,signature,exceptions,type, this.castRequired, this.adapterMethod
                                     ));
                             }
                         };
